@@ -22,7 +22,6 @@
 
 #include "config.h"
 
-#include <dazzle.h>
 #include <glib/gi18n.h>
 #include <stdlib.h>
 
@@ -44,7 +43,7 @@ struct _IdeExtensionSetAdapter
   guint       reload_handler;
 };
 
-G_DEFINE_TYPE (IdeExtensionSetAdapter, ide_extension_set_adapter, IDE_TYPE_OBJECT)
+G_DEFINE_FINAL_TYPE (IdeExtensionSetAdapter, ide_extension_set_adapter, IDE_TYPE_OBJECT)
 
 enum {
   EXTENSIONS_LOADED,
@@ -150,8 +149,8 @@ watch_extension (IdeExtensionSetAdapter *self,
                  PeasPluginInfo         *plugin_info,
                  GType                   interface_type)
 {
-  GSettings *settings;
-  gchar *path;
+  g_autoptr(GSettings) settings = NULL;
+  g_autofree char *path = NULL;
 
   g_assert (IDE_IS_MAIN_THREAD ());
   g_assert (IDE_IS_EXTENSION_SET_ADAPTER (self));
@@ -165,14 +164,14 @@ watch_extension (IdeExtensionSetAdapter *self,
 
   g_ptr_array_add (self->settings, g_object_ref (settings));
 
+  /* We have to fetch the key once to get changed events */
+  g_settings_get_boolean (settings, "enabled");
+
   g_signal_connect_object (settings,
                            "changed::enabled",
                            G_CALLBACK (ide_extension_set_adapter_enabled_changed),
                            self,
                            G_CONNECT_SWAPPED);
-
-  g_object_unref (settings);
-  g_free (path);
 }
 
 static void
@@ -553,8 +552,6 @@ ide_extension_set_adapter_init (IdeExtensionSetAdapter *self)
  * Gets the #IdeExtensionSetAdapter:engine property.
  *
  * Returns: (transfer none): a #PeasEngine.
- *
- * Since: 3.32
  */
 PeasEngine *
 ide_extension_set_adapter_get_engine (IdeExtensionSetAdapter *self)
@@ -632,8 +629,6 @@ ide_extension_set_adapter_set_value (IdeExtensionSetAdapter *self,
  * @user_data: user data for @foreach_func
  *
  * Calls @foreach_func for every extension loaded by the extension set.
- *
- * Since: 3.32
  */
 void
 ide_extension_set_adapter_foreach (IdeExtensionSetAdapter            *self,
@@ -693,8 +688,6 @@ sort_by_priority (gconstpointer a,
  * @user_data: user data for @foreach_func
  *
  * Calls @foreach_func for every extension loaded by the extension set.
- *
- * Since: 3.32
  */
 void
 ide_extension_set_adapter_foreach_by_priority (IdeExtensionSetAdapter            *self,
@@ -799,8 +792,6 @@ ide_extension_set_adapter_new (IdeObject   *parent,
  * Locates the extension owned by @plugin_info if such extension exists.
  *
  * Returns: (transfer none) (nullable): a #PeasExtension or %NULL
- *
- * Since: 3.32
  */
 PeasExtension *
 ide_extension_set_adapter_get_extension (IdeExtensionSetAdapter *self,
