@@ -30,8 +30,9 @@
 #include "ide-search-engine.h"
 #include "ide-search-provider.h"
 #include "ide-search-result.h"
+#include "ide-search-results-private.h"
 
-#define DEFAULT_MAX_RESULTS 50
+#define DEFAULT_MAX_RESULTS 100
 
 struct _IdeSearchEngine
 {
@@ -43,11 +44,18 @@ struct _IdeSearchEngine
 
 typedef struct
 {
-  IdeTask       *task;
-  gchar         *query;
-  GListStore    *store;
-  guint          outstanding;
-  guint          max_results;
+  IdeSearchProvider *provider;
+  GListModel        *results;
+  guint              truncated : 1;
+} SortInfo;
+
+typedef struct
+{
+  IdeTask    *task;
+  char       *query;
+  GArray     *sorted;
+  guint       outstanding;
+  guint       max_results;
 } Request;
 
 enum {
@@ -56,19 +64,29 @@ enum {
   N_PROPS
 };
 
-G_DEFINE_TYPE (IdeSearchEngine, ide_search_engine, IDE_TYPE_OBJECT)
+G_DEFINE_FINAL_TYPE (IdeSearchEngine, ide_search_engine, IDE_TYPE_OBJECT)
 
 static GParamSpec *properties [N_PROPS];
 
+static void
+sort_info_clear (gpointer item)
+{
+  SortInfo *info = item;
+
+  g_clear_object (&info->provider);
+  g_clear_object (&info->results);
+}
+
 static Request *
 request_new (void)
 {
   Request *r;
 
   r = g_slice_new0 (Request);
-  r->store = NULL;
   r->outstanding = 0;
   r->query = NULL;
+  r->sorted = g_array_new (FALSE, FALSE, sizeof (SortInfo));
+  g_array_set_clear_func (r->sorted, sort_info_clear);
 
   return r;
 }
@@ -77,8 +95,8 @@ static void
 request_destroy (Request *r)
 {
   g_assert (r->outstanding == 0);
-  g_clear_object (&r->store);
   g_clear_pointer (&r->query, g_free);
+  g_clear_pointer (&r->sorted, g_array_unref);
   r->task = NULL;
   g_slice_free (Request, r);
 }
@@ -89,17 +107,13 @@ on_extension_added_cb (IdeExtensionSetAdapter *set,
                        PeasExtension          *exten,
                        gpointer                user_data)
 {
-  IdeSearchEngine *self = (IdeSearchEngine *)user_data;
   IdeSearchProvider *provider = (IdeSearchProvider *)exten;
-  IdeContext *context = ide_object_get_context (IDE_OBJECT (self));
 
   g_assert (IDE_IS_MAIN_THREAD ());
   g_assert (plugin_info != NULL);
   g_assert (IDE_IS_SEARCH_PROVIDER (provider));
-  g_assert (context != NULL);
-  g_assert (IDE_IS_CONTEXT (context));
 
-  ide_search_provider_load (provider, context);
+  ide_search_provider_load (provider);
 }
 
 static void
@@ -108,7 +122,13 @@ on_extension_removed_cb (IdeExtensionSetAdapter *set,
                          PeasExtension          *exten,
                          gpointer                user_data)
 {
-// FIXME ??
+  IdeSearchProvider *provider = (IdeSearchProvider *)exten;
+
+  g_assert (IDE_IS_MAIN_THREAD ());
+  g_assert (plugin_info != NULL);
+  g_assert (IDE_IS_SEARCH_PROVIDER (provider));
+
+  ide_search_provider_unload (provider);
 }
 
 static void
@@ -121,15 +141,12 @@ ide_search_engine_parent_set (IdeObject *object,
   g_assert (!parent || IDE_IS_OBJECT (parent));
 
   if (parent == NULL)
-    {
-      g_clear_object (&self->extensions);
-      return;
-    }
+    return;
 
-  self->extensions = ide_extension_set_adapter_new (parent,
+  self->extensions = ide_extension_set_adapter_new (object,
                                                     peas_engine_get_default (),
                                                     IDE_TYPE_SEARCH_PROVIDER,
-                                                    NULL, NULL);
+                                                    "Search-Provider", NULL);
 
   g_signal_connect (self->extensions,
                     "extension-added",
@@ -147,14 +164,14 @@ ide_search_engine_parent_set (IdeObject *object,
 }
 
 static void
-ide_search_engine_destroy (IdeObject *object)
+ide_search_engine_dispose (GObject *object)
 {
   IdeSearchEngine *self = (IdeSearchEngine *)object;
 
   g_clear_object (&self->extensions);
   g_clear_pointer (&self->custom_provider, g_ptr_array_unref);
 
-  IDE_OBJECT_CLASS (ide_search_engine_parent_class)->destroy (object);
+  G_OBJECT_CLASS (ide_search_engine_parent_class)->dispose (object);
 }
 
 static void
@@ -179,13 +196,13 @@ ide_search_engine_get_property (GObject    *object,
 static void
 ide_search_engine_class_init (IdeSearchEngineClass *klass)
 {
-  GObjectClass *g_object_class = G_OBJECT_CLASS (klass);
-  IdeObjectClass *object_class = IDE_OBJECT_CLASS (klass);
+  GObjectClass *object_class = G_OBJECT_CLASS (klass);
+  IdeObjectClass *i_object_class = IDE_OBJECT_CLASS (klass);
 
-  g_object_class->get_property = ide_search_engine_get_property;
+  object_class->dispose = ide_search_engine_dispose;
+  object_class->get_property = ide_search_engine_get_property;
 
-  object_class->destroy = ide_search_engine_destroy;
-  object_class->parent_set = ide_search_engine_parent_set;
+  i_object_class->parent_set = ide_search_engine_parent_set;
 
   properties [PROP_BUSY] =
     g_param_spec_boolean ("busy",
@@ -194,7 +211,7 @@ ide_search_engine_class_init (IdeSearchEngineClass *klass)
                           FALSE,
                           (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 
-  g_object_class_install_properties (g_object_class, N_PROPS, properties);
+  g_object_class_install_properties (object_class, N_PROPS, properties);
 }
 
 static void
@@ -216,8 +233,6 @@ ide_search_engine_new (void)
  * Checks if the #IdeSearchEngine is currently executing a query.
  *
  * Returns: %TRUE if queries are being processed.
- *
- * Since: 3.32
  */
 gboolean
 ide_search_engine_get_busy (IdeSearchEngine *self)
@@ -235,7 +250,7 @@ ide_search_engine_search_cb (GObject      *object,
   IdeSearchProvider *provider = (IdeSearchProvider *)object;
   g_autoptr(IdeTask) task = user_data;
   g_autoptr(GError) error = NULL;
-  g_autoptr(GPtrArray) ar = NULL;
+  g_autoptr(GListModel) ret = NULL;
   Request *r;
 
   g_assert (IDE_IS_SEARCH_PROVIDER (provider));
@@ -246,58 +261,84 @@ ide_search_engine_search_cb (GObject      *object,
   g_assert (r != NULL);
   g_assert (r->task == task);
   g_assert (r->outstanding > 0);
-  g_assert (G_IS_LIST_STORE (r->store));
 
-  ar = ide_search_provider_search_finish (provider, result, &error);
-  IDE_PTR_ARRAY_SET_FREE_FUNC (ar, g_object_unref);
-
-  if (error != NULL)
+  for (guint i = 0; i < r->sorted->len; i++)
     {
-      if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED) &&
-          !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED))
-        g_warning ("%s", error->message);
-      goto cleanup;
+      SortInfo *info = &g_array_index (r->sorted, SortInfo, i);
+      gboolean truncated = FALSE;
+
+      if (info->provider != provider)
+        continue;
+
+      g_assert (info->results == NULL);
+
+      if (!(info->results = ide_search_provider_search_finish (provider, result, &truncated, &error)))
+        {
+          IDE_TRACE_MSG ("%s: %s", G_OBJECT_TYPE_NAME (provider), error->message);
+
+          if (!ide_error_ignore (error))
+            g_warning ("%s", error->message);
+        }
+
+      info->truncated = info->results != NULL && truncated;
+
+#ifdef IDE_ENABLE_TRACE
+      if (info->results != NULL)
+        IDE_TRACE_MSG ("%s: %d results%s",
+                       G_OBJECT_TYPE_NAME (provider),
+                       g_list_model_get_n_items (info->results),
+                       info->truncated ? " [truncated]" : "");
+#endif
+
+      break;
     }
 
-  for (guint i = 0; i < ar->len; i++)
-    {
-      IdeSearchResult *item = g_ptr_array_index (ar, i);
-
-      g_assert (IDE_IS_SEARCH_RESULT (item));
-
-      g_list_store_insert_sorted (r->store,
-                                  item,
-                                  (GCompareDataFunc)ide_search_result_compare,
-                                  NULL);
-
-    }
-
-cleanup:
   r->outstanding--;
 
   if (r->outstanding == 0)
-    ide_task_return_pointer (task, g_steal_pointer (&r->store), g_object_unref);
+    {
+      g_autoptr(GListStore) store = g_list_store_new (G_TYPE_LIST_MODEL);
+      g_autoptr(GtkFlattenListModel) flatten = gtk_flatten_list_model_new (G_LIST_MODEL (g_object_ref (store)));
+      gboolean truncated = FALSE;
+
+      for (guint i = 0; i < r->sorted->len; i++)
+        {
+          SortInfo *info = &g_array_index (r->sorted, SortInfo, i);
+
+          if (info->results != NULL)
+            g_list_store_append (store, info->results);
+
+          truncated |= info->truncated;
+        }
+
+      ide_task_return_pointer (task,
+                               _ide_search_results_new (G_LIST_MODEL (flatten), r->query, truncated),
+                               g_object_unref);
+    }
 }
 
 static void
 _provider_search_async (IdeSearchProvider *provider,
-                        Request           *request)
+                        Request           *r)
 {
-  g_assert (IDE_IS_SEARCH_PROVIDER (provider));
-  g_assert (request != NULL);
-  g_assert (IDE_IS_TASK (request->task));
-  g_assert (G_IS_LIST_STORE (request->store));
+  SortInfo sort_info;
 
-  request->outstanding++;
+  g_assert (IDE_IS_SEARCH_PROVIDER (provider));
+  g_assert (r != NULL);
+  g_assert (IDE_IS_TASK (r->task));
+
+  r->outstanding++;
+
+  sort_info.provider = g_object_ref (provider);
+  sort_info.results = NULL;
+  g_array_append_val (r->sorted, sort_info);
 
   ide_search_provider_search_async (provider,
-                                    request->query,
-                                    request->max_results,
-                                    ide_task_get_cancellable (request->task),
+                                    r->query,
+                                    r->max_results,
+                                    ide_task_get_cancellable (r->task),
                                     ide_search_engine_search_cb,
-                                    g_object_ref (request->task));
-
-
+                                    g_object_ref (r->task));
 }
 
 static void
@@ -314,7 +355,6 @@ ide_search_engine_search_foreach (IdeExtensionSetAdapter *set,
   g_assert (IDE_IS_SEARCH_PROVIDER (provider));
   g_assert (r != NULL);
   g_assert (IDE_IS_TASK (r->task));
-  g_assert (G_IS_LIST_STORE (r->store));
 
   _provider_search_async (provider, r);
 }
@@ -329,14 +369,13 @@ ide_search_engine_search_foreach_custom_provider (gpointer data,
   g_assert (IDE_IS_SEARCH_PROVIDER (provider));
   g_assert (r != NULL);
   g_assert (IDE_IS_TASK (r->task));
-  g_assert (G_IS_LIST_STORE (r->store));
 
   _provider_search_async (provider, r);
 }
 
 void
 ide_search_engine_search_async (IdeSearchEngine     *self,
-                                const gchar         *query,
+                                const char          *query,
                                 guint                max_results,
                                 GCancellable        *cancellable,
                                 GAsyncReadyCallback  callback,
@@ -359,23 +398,20 @@ ide_search_engine_search_async (IdeSearchEngine     *self,
   r->query = g_strdup (query);
   r->max_results = max_results;
   r->task = task;
-  r->store = g_list_store_new (IDE_TYPE_SEARCH_RESULT);
   r->outstanding = 0;
   ide_task_set_task_data (task, r, request_destroy);
 
-  ide_extension_set_adapter_foreach (self->extensions,
-                                     ide_search_engine_search_foreach,
-                                     r);
   g_ptr_array_foreach (self->custom_provider,
                        ide_search_engine_search_foreach_custom_provider,
                        r);
+  ide_extension_set_adapter_foreach_by_priority (self->extensions,
+                                                 ide_search_engine_search_foreach,
+                                                 r);
 
   self->active_count += r->outstanding;
 
   if (r->outstanding == 0)
-    ide_task_return_pointer (task,
-                             g_object_ref (r->store),
-                             g_object_unref);
+    ide_task_return_unsupported_error (task);
 
   g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_BUSY]);
 }
@@ -388,21 +424,29 @@ ide_search_engine_search_async (IdeSearchEngine     *self,
  *
  * Completes an asynchronous request to ide_search_engine_search_async().
  *
- * The result is a #GListModel of #IdeSearchResult when successful.
+ * The result is a #GListModel of #IdeSearchResult when successful. The type
+ * is #IdeSearchResults which allows you to do additional filtering on the
+ * result set instead of querying providers again.
  *
  * Returns: (transfer full): a #GListModel of #IdeSearchResult items.
- *
- * Since: 3.32
  */
-GListModel *
+IdeSearchResults *
 ide_search_engine_search_finish (IdeSearchEngine  *self,
                                  GAsyncResult     *result,
                                  GError          **error)
 {
+  IdeSearchResults *ret;
+
+  IDE_ENTRY;
+
   g_return_val_if_fail (IDE_IS_SEARCH_ENGINE (self), NULL);
   g_return_val_if_fail (IDE_IS_TASK (result), NULL);
 
-  return ide_task_propagate_pointer (IDE_TASK (result), error);
+  ret = ide_task_propagate_pointer (IDE_TASK (result), error);
+
+  g_return_val_if_fail (!ret || IDE_IS_SEARCH_RESULTS (ret), NULL);
+
+  IDE_RETURN (ret);
 }
 
 /**
@@ -412,8 +456,6 @@ ide_search_engine_search_finish (IdeSearchEngine  *self,
  *
  * Adds a custom search provider to the #IdeSearchEngine. This enables
  * to bring in custom #IdeSearchProvider during the runtime.
- *
- * Since: 3.38
  */
 void
 ide_search_engine_add_provider (IdeSearchEngine   *self,
@@ -432,8 +474,6 @@ ide_search_engine_add_provider (IdeSearchEngine   *self,
  *
  * Remove a custom search provider from the #IdeSearchEngine. This removes
  * custom #IdeSearchProvider during the runtime.
- *
- * Since: 3.38
  */
 void
 ide_search_engine_remove_provider (IdeSearchEngine   *self,
