@@ -22,21 +22,21 @@
 
 #include "config.h"
 
-#include <fcntl.h>
 #include <glib/gi18n.h>
+
+#include <fcntl.h>
+#include <stdlib.h>
+#include <unistd.h>
+#include <vte/vte.h>
+
 #include <libide-foundry.h>
 #include <libide-gui.h>
 #include <libide-terminal.h>
-#include <stdlib.h>
-#include <vte/vte.h>
-#include <unistd.h>
-
-#define PCRE2_CODE_UNIT_WIDTH 0
-#include <pcre2.h>
 
 #include "ide-terminal-page.h"
 #include "ide-terminal-page-private.h"
 #include "ide-terminal-page-actions.h"
+#include "ide-terminal-run-command-private.h"
 
 #define FLAPPING_DURATION_USEC (G_USEC_PER_SEC / 20)
 
@@ -52,13 +52,7 @@ enum {
   N_PROPS
 };
 
-enum {
-  TEXT_INSERTED,
-  N_SIGNALS
-};
-
 static GParamSpec *properties [N_PROPS];
-static guint signals [N_SIGNALS];
 
 static void ide_terminal_page_connect_terminal (IdeTerminalPage *self,
                                                 VteTerminal     *terminal);
@@ -79,11 +73,14 @@ terminal_has_notification_signal (void)
 static gboolean
 destroy_widget_in_idle (GtkWidget *widget)
 {
+  IdeTerminalPage *self = (IdeTerminalPage *)widget;
+
   IDE_ENTRY;
 
-  g_assert (GTK_IS_WIDGET (widget));
+  g_assert (IDE_IS_TERMINAL_PAGE (self));
 
-  gtk_widget_destroy (widget);
+  if (!self->destroyed)
+    panel_widget_close (PANEL_WIDGET (self));
 
   IDE_RETURN (G_SOURCE_REMOVE);
 }
@@ -108,13 +105,12 @@ ide_terminal_page_spawn_cb (GObject      *object,
 
   self->exited = TRUE;
 
-  title = g_strdup_printf ("%s (%s)",
-                           ide_page_get_title (IDE_PAGE (self)) ?: _("Untitled terminal"),
-                           /* translators: exited describes that the terminal shell process has exited */
-                           _("Exited"));
-  ide_page_set_title (IDE_PAGE (self), title);
+  ide_terminal_launcher_spawn_finish (launcher, result, &error);
 
-  if (!ide_terminal_launcher_spawn_finish (launcher, result, &error))
+  if (self->destroyed)
+    IDE_EXIT;
+
+  if (error != NULL)
     {
       g_autofree gchar *format = NULL;
 
@@ -122,8 +118,11 @@ ide_terminal_page_spawn_cb (GObject      *object,
       ide_terminal_page_feed (self, format);
     }
 
-  if (gtk_widget_in_destruction (GTK_WIDGET (self)))
-    IDE_EXIT;
+  title = g_strdup_printf ("%s (%s)",
+                           panel_widget_get_title (PANEL_WIDGET (self)) ?: _("Untitled terminal"),
+                           /* translators: exited describes that the terminal shell process has exited */
+                           _("Exited"));
+  panel_widget_set_title (PANEL_WIDGET (self), title);
 
   now = g_get_monotonic_time ();
   maybe_flapping = ABS (now - self->last_respawn) < FLAPPING_DURATION_USEC;
@@ -131,12 +130,12 @@ ide_terminal_page_spawn_cb (GObject      *object,
   if (!self->respawn_on_exit)
     {
       if (self->close_on_exit && !maybe_flapping)
-        gdk_threads_add_idle_full (G_PRIORITY_LOW + 1000,
-                                   (GSourceFunc) destroy_widget_in_idle,
-                                   g_object_ref (self),
-                                   g_object_unref);
+        g_idle_add_full (G_PRIORITY_LOW + 1000,
+                         (GSourceFunc) destroy_widget_in_idle,
+                         g_object_ref (self),
+                         g_object_unref);
       else
-        vte_terminal_set_input_enabled (VTE_TERMINAL (self->terminal_top), FALSE);
+        vte_terminal_set_input_enabled (VTE_TERMINAL (self->terminal), FALSE);
       IDE_EXIT;
     }
 
@@ -148,14 +147,14 @@ ide_terminal_page_spawn_cb (GObject      *object,
     }
 
   g_clear_object (&self->pty);
-  vte_terminal_reset (VTE_TERMINAL (self->terminal_top), TRUE, TRUE);
-  self->pty = vte_pty_new_sync (VTE_PTY_DEFAULT, NULL, NULL);
-  vte_terminal_set_pty (VTE_TERMINAL (self->terminal_top), self->pty);
+  vte_terminal_reset (VTE_TERMINAL (self->terminal), TRUE, TRUE);
+  self->pty = ide_pty_new_sync (NULL);
+  vte_terminal_set_pty (VTE_TERMINAL (self->terminal), self->pty);
 
   /* Spawn our terminal and wait for it to exit */
   self->last_respawn = now;
   self->exited = FALSE;
-  ide_page_set_title (IDE_PAGE (self), _("Untitled terminal"));
+  panel_widget_set_title (PANEL_WIDGET (self), _("Untitled terminal"));
   ide_terminal_launcher_spawn_async (self->launcher,
                                      self->pty,
                                      NULL,
@@ -165,8 +164,46 @@ ide_terminal_page_spawn_cb (GObject      *object,
   IDE_EXIT;
 }
 
+static gboolean
+ide_terminal_page_do_spawn_in_idle (IdeTerminalPage *self)
+{
+  IDE_ENTRY;
+
+  g_assert (IDE_IS_TERMINAL_PAGE (self));
+
+  if (self->destroyed)
+    IDE_RETURN (G_SOURCE_REMOVE);
+
+  self->last_respawn = g_get_monotonic_time ();
+
+  if (self->pty == NULL)
+    {
+      g_autoptr(GError) error = NULL;
+
+      if (!(self->pty = ide_pty_new_sync (&error)))
+        {
+          g_critical ("Failed to create PTY for terminal: %s", error->message);
+          IDE_RETURN (G_SOURCE_REMOVE);
+        }
+    }
+
+  vte_terminal_set_pty (VTE_TERMINAL (self->terminal), self->pty);
+
+  if (!self->manage_spawn)
+    IDE_RETURN (G_SOURCE_REMOVE);
+
+  /* Spawn our terminal and wait for it to exit */
+  ide_terminal_launcher_spawn_async (self->launcher,
+                                     self->pty,
+                                     NULL,
+                                     ide_terminal_page_spawn_cb,
+                                     g_object_ref (self));
+
+  IDE_RETURN (G_SOURCE_REMOVE);
+}
+
 static void
-gbp_terminal_page_realize (GtkWidget *widget)
+ide_terminal_page_realize (GtkWidget *widget)
 {
   IdeTerminalPage *self = (IdeTerminalPage *)widget;
 
@@ -179,83 +216,14 @@ gbp_terminal_page_realize (GtkWidget *widget)
 
   self->did_defered_setup_in_realize = TRUE;
 
-  self->last_respawn = g_get_monotonic_time ();
-
-  if (self->pty == NULL)
-    {
-      g_autoptr(GError) error = NULL;
-
-      if (!(self->pty = vte_pty_new_sync (VTE_PTY_DEFAULT, NULL, &error)))
-        {
-          g_critical ("Failed to create PTY for terminal: %s", error->message);
-          return;
-        }
-    }
-
-  vte_terminal_set_pty (VTE_TERMINAL (self->terminal_top), self->pty);
-
-  if (!self->manage_spawn)
-    return;
-
-  /* Spawn our terminal and wait for it to exit */
-  ide_terminal_launcher_spawn_async (self->launcher,
-                                     self->pty,
-                                     NULL,
-                                     ide_terminal_page_spawn_cb,
-                                     g_object_ref (self));
-}
-
-static void
-gbp_terminal_page_get_preferred_width (GtkWidget *widget,
-                                       gint      *min_width,
-                                       gint      *nat_width)
-{
-  /*
-   * Since we are placing the terminal in a GtkStack, we need
-   * to fake the size a bit. Otherwise, GtkStack tries to keep the
-   * widget at its natural size (which prevents us from getting
-   * appropriate size requests.
+  /* We don't want to process this in realize as it could be holding things
+   * up from being mapped. Instead, wait until the GDK backend has finished
+   * reacting to realize/etc and then spawn from idle.
    */
-  GTK_WIDGET_CLASS (ide_terminal_page_parent_class)->get_preferred_width (widget, min_width, nat_width);
-  *nat_width = *min_width;
-}
-
-static void
-gbp_terminal_page_get_preferred_height (GtkWidget *widget,
-                                        gint      *min_height,
-                                        gint      *nat_height)
-{
-  /*
-   * Since we are placing the terminal in a GtkStack, we need
-   * to fake the size a bit. Otherwise, GtkStack tries to keep the
-   * widget at its natural size (which prevents us from getting
-   * appropriate size requests.
-   */
-  GTK_WIDGET_CLASS (ide_terminal_page_parent_class)->get_preferred_height (widget, min_height, nat_height);
-  *nat_height = *min_height;
-}
-
-static void
-gbp_terminal_page_set_needs_attention (IdeTerminalPage *self,
-                                       gboolean         needs_attention)
-{
-  GtkWidget *parent;
-
-  g_assert (IDE_IS_TERMINAL_PAGE (self));
-
-  parent = gtk_widget_get_parent (GTK_WIDGET (self));
-
-  if (GTK_IS_STACK (parent) &&
-      !gtk_widget_in_destruction (GTK_WIDGET (self)) &&
-      !gtk_widget_in_destruction (parent))
-    {
-      if (!gtk_widget_in_destruction (GTK_WIDGET (self->terminal_top)))
-        self->needs_attention = !!needs_attention;
-
-      gtk_container_child_set (GTK_CONTAINER (parent), GTK_WIDGET (self),
-                               "needs-attention", needs_attention,
-                               NULL);
-    }
+  g_idle_add_full (G_PRIORITY_LOW,
+                   (GSourceFunc)ide_terminal_page_do_spawn_in_idle,
+                   g_object_ref (self),
+                   g_object_unref);
 }
 
 static void
@@ -267,23 +235,22 @@ notification_received_cb (VteTerminal     *terminal,
   g_assert (VTE_IS_TERMINAL (terminal));
   g_assert (IDE_IS_TERMINAL_PAGE (self));
 
+  if (self->destroyed)
+    return;
+
   if (!gtk_widget_has_focus (GTK_WIDGET (terminal)))
-    gbp_terminal_page_set_needs_attention (self, TRUE);
+    panel_widget_set_needs_attention (PANEL_WIDGET (self), TRUE);
 }
 
-static gboolean
-focus_in_event_cb (VteTerminal     *terminal,
-                   GdkEvent        *event,
-                   IdeTerminalPage *self)
+static void
+focus_in_event_cb (GtkEventControllerFocus *focus,
+                   IdeTerminalPage         *self)
 {
-  g_assert (VTE_IS_TERMINAL (terminal));
+  g_assert (GTK_IS_EVENT_CONTROLLER_FOCUS (focus));
   g_assert (IDE_IS_TERMINAL_PAGE (self));
 
-  self->needs_attention = FALSE;
-  gbp_terminal_page_set_needs_attention (self, FALSE);
-  gtk_revealer_set_reveal_child (self->search_revealer_top, FALSE);
-
-  return GDK_EVENT_PROPAGATE;
+  panel_widget_set_needs_attention (PANEL_WIDGET (self), FALSE);
+  gtk_revealer_set_reveal_child (self->search_revealer, FALSE);
 }
 
 static void
@@ -295,41 +262,19 @@ window_title_changed_cb (VteTerminal     *terminal,
   g_assert (VTE_IS_TERMINAL (terminal));
   g_assert (IDE_IS_TERMINAL_PAGE (self));
 
-  title = vte_terminal_get_window_title (VTE_TERMINAL (self->terminal_top));
+  if (self->destroyed)
+    return;
+
+  title = vte_terminal_get_window_title (VTE_TERMINAL (self->terminal));
 
   if (title == NULL || title[0] == '\0')
     title = _("Untitled terminal");
 
-  ide_page_set_title (IDE_PAGE (self), title);
-}
-
-static void
-style_context_changed (GtkStyleContext *style_context,
-                       IdeTerminalPage *self)
-{
-  GtkStateFlags state;
-  GdkRGBA fg;
-  GdkRGBA bg;
-
-  g_assert (GTK_IS_STYLE_CONTEXT (style_context));
-  g_assert (IDE_IS_TERMINAL_PAGE (self));
-
-  state = gtk_style_context_get_state (style_context);
-
-  G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
-  gtk_style_context_get_color (style_context, state, &fg);
-  gtk_style_context_get_background_color (style_context, state, &bg);
-  G_GNUC_END_IGNORE_DEPRECATIONS;
-
-  if (bg.alpha == 0.0)
-    gdk_rgba_parse (&bg, "#f6f7f8");
-
-  ide_page_set_primary_color_fg (IDE_PAGE (self), &fg);
-  ide_page_set_primary_color_bg (IDE_PAGE (self), &bg);
+  panel_widget_set_title (PANEL_WIDGET (self), title);
 }
 
 static IdePage *
-gbp_terminal_page_create_split (IdePage *page)
+ide_terminal_page_create_split (IdePage *page)
 {
   IdeTerminalPage *self = (IdeTerminalPage *)page;
 
@@ -339,37 +284,42 @@ gbp_terminal_page_create_split (IdePage *page)
                        "close-on-exit", self->close_on_exit,
                        "launcher", self->launcher,
                        "manage-spawn", self->manage_spawn,
-                       "pty", self->manage_spawn ? NULL : self->pty,
+                       "pty", NULL,
                        "respawn-on-exit", self->respawn_on_exit,
                        "visible", TRUE,
                        NULL);
 }
 
-static void
-gbp_terminal_page_grab_focus (GtkWidget *widget)
+static gboolean
+ide_terminal_page_grab_focus (GtkWidget *widget)
 {
   IdeTerminalPage *self = (IdeTerminalPage *)widget;
 
   g_assert (IDE_IS_TERMINAL_PAGE (self));
 
-  gtk_widget_grab_focus (GTK_WIDGET (self->terminal_top));
+  return gtk_widget_grab_focus (GTK_WIDGET (self->terminal));
 }
 
 static void
 ide_terminal_page_connect_terminal (IdeTerminalPage *self,
                                     VteTerminal     *terminal)
 {
-  GtkAdjustment *vadj;
+  GtkEventController *controller;
 
-  vadj = gtk_scrollable_get_vadjustment (GTK_SCROLLABLE (terminal));
+  g_assert (IDE_IS_TERMINAL_PAGE (self));
+  g_assert (IDE_IS_TERMINAL (terminal));
 
-  gtk_range_set_adjustment (GTK_RANGE (self->top_scrollbar), vadj);
+  if (self->destroyed)
+    return;
 
-  g_signal_connect_object (terminal,
-                           "focus-in-event",
+  controller = gtk_event_controller_focus_new ();
+  g_signal_connect_object (controller,
+                           "enter",
                            G_CALLBACK (focus_in_event_cb),
                            self,
                            0);
+  gtk_widget_add_controller (GTK_WIDGET (terminal), controller);
+
 
   g_signal_connect_object (terminal,
                            "window-title-changed",
@@ -396,15 +346,14 @@ ide_terminal_page_context_set (GtkWidget  *widget,
   g_assert (IDE_IS_TERMINAL_PAGE (self));
   g_assert (!context || IDE_IS_CONTEXT (context));
 
-  if (self->launcher == NULL && context != NULL)
-    self->launcher = ide_terminal_launcher_new (context);
-}
+  if (context == NULL)
+    return;
 
-static void
-ide_terminal_page_on_text_inserted_cb (IdeTerminalPage *self,
-                                       VteTerminal     *terminal)
-{
-  g_signal_emit (self, signals [TEXT_INSERTED], 0);
+  if (self->launcher == NULL)
+    {
+      g_autoptr(IdeRunCommand) run_command = ide_terminal_run_command_new (IDE_TERMINAL_RUN_ON_HOST);
+      self->launcher = ide_terminal_launcher_new (context, run_command);
+    }
 }
 
 static GFile *
@@ -415,8 +364,11 @@ ide_terminal_page_get_file_or_directory (IdePage *page)
 
   g_assert (IDE_IS_TERMINAL_PAGE (self));
 
-  if (!(uri = vte_terminal_get_current_file_uri (VTE_TERMINAL (self->terminal_top))))
-    uri = vte_terminal_get_current_directory_uri (VTE_TERMINAL (self->terminal_top));
+  if (self->destroyed)
+    return NULL;
+
+  if (!(uri = vte_terminal_get_current_file_uri (VTE_TERMINAL (self->terminal))))
+    uri = vte_terminal_get_current_directory_uri (VTE_TERMINAL (self->terminal));
 
   if (uri != NULL)
     return g_file_new_for_uri (uri);
@@ -425,16 +377,18 @@ ide_terminal_page_get_file_or_directory (IdePage *page)
 }
 
 static void
-ide_terminal_page_finalize (GObject *object)
+ide_terminal_page_dispose (GObject *object)
 {
   IdeTerminalPage *self = IDE_TERMINAL_PAGE (object);
 
+  self->destroyed = TRUE;
+
   g_clear_object (&self->launcher);
-  g_clear_object (&self->save_as_file_top);
+  g_clear_object (&self->save_as_file);
   g_clear_pointer (&self->selection_buffer, g_free);
   g_clear_object (&self->pty);
 
-  G_OBJECT_CLASS (ide_terminal_page_parent_class)->finalize (object);
+  G_OBJECT_CLASS (ide_terminal_page_parent_class)->dispose (object);
 }
 
 static void
@@ -514,22 +468,19 @@ ide_terminal_page_class_init (IdeTerminalPageClass *klass)
   GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
   IdePageClass *page_class = IDE_PAGE_CLASS (klass);
 
-  object_class->finalize = ide_terminal_page_finalize;
+  object_class->dispose = ide_terminal_page_dispose;
   object_class->get_property = ide_terminal_page_get_property;
   object_class->set_property = ide_terminal_page_set_property;
 
-  widget_class->realize = gbp_terminal_page_realize;
-  widget_class->get_preferred_width = gbp_terminal_page_get_preferred_width;
-  widget_class->get_preferred_height = gbp_terminal_page_get_preferred_height;
-  widget_class->grab_focus = gbp_terminal_page_grab_focus;
+  widget_class->realize = ide_terminal_page_realize;
+  widget_class->grab_focus = ide_terminal_page_grab_focus;
 
-  page_class->create_split = gbp_terminal_page_create_split;
+  page_class->create_split = ide_terminal_page_create_split;
   page_class->get_file_or_directory = ide_terminal_page_get_file_or_directory;
 
   gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/libide-terminal/ui/ide-terminal-page.ui");
-  gtk_widget_class_bind_template_child (widget_class, IdeTerminalPage, terminal_top);
-  gtk_widget_class_bind_template_child (widget_class, IdeTerminalPage, top_scrollbar);
-  gtk_widget_class_bind_template_child (widget_class, IdeTerminalPage, terminal_overlay_top);
+  gtk_widget_class_bind_template_child (widget_class, IdeTerminalPage, terminal);
+  gtk_widget_class_bind_template_child (widget_class, IdeTerminalPage, terminal_overlay);
 
   properties [PROP_CLOSE_ON_EXIT] =
     g_param_spec_boolean ("close-on-exit",
@@ -564,28 +515,14 @@ ide_terminal_page_class_init (IdeTerminalPageClass *klass)
                          "Launcher",
                          "The launcher to use for spawning",
                          IDE_TYPE_TERMINAL_LAUNCHER,
-                         (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+                         (G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS));
 
   g_object_class_install_properties (object_class, N_PROPS, properties);
-
-  signals [TEXT_INSERTED] =
-    g_signal_new ("text-inserted",
-                  G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST,
-                  0,
-                  NULL, NULL,
-                  g_cclosure_marshal_VOID__VOID,
-                  G_TYPE_NONE, 0);
-  g_signal_set_va_marshaller (signals [TEXT_INSERTED],
-                              G_TYPE_FROM_CLASS (klass),
-                              g_cclosure_marshal_VOID__VOIDv);
 }
 
 static void
 ide_terminal_page_init (IdeTerminalPage *self)
 {
-  GtkStyleContext *style_context;
-
   self->close_on_exit = TRUE;
   self->respawn_on_exit = TRUE;
   self->manage_spawn = TRUE;
@@ -593,42 +530,41 @@ ide_terminal_page_init (IdeTerminalPage *self)
   self->tsearch = g_object_new (IDE_TYPE_TERMINAL_SEARCH,
                                 "visible", TRUE,
                                 NULL);
-  self->search_revealer_top = ide_terminal_search_get_revealer (self->tsearch);
+  self->search_revealer = ide_terminal_search_get_revealer (self->tsearch);
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
-  g_signal_connect_object (self->terminal_top,
-                           "text-inserted",
-                           G_CALLBACK (ide_terminal_page_on_text_inserted_cb),
-                           self,
-                           G_CONNECT_SWAPPED);
-
-  ide_page_set_icon_name (IDE_PAGE (self), "utilities-terminal-symbolic");
+  panel_widget_set_icon_name (PANEL_WIDGET (self), "builder-terminal-symbolic");
   ide_page_set_can_split (IDE_PAGE (self), TRUE);
   ide_page_set_menu_id (IDE_PAGE (self), "ide-terminal-page-document-menu");
 
-  gtk_overlay_add_overlay (self->terminal_overlay_top, GTK_WIDGET (self->tsearch));
+  gtk_overlay_add_overlay (self->terminal_overlay, GTK_WIDGET (self->tsearch));
 
-  ide_terminal_page_connect_terminal (self, VTE_TERMINAL (self->terminal_top));
+  ide_terminal_page_connect_terminal (self, VTE_TERMINAL (self->terminal));
 
-  ide_terminal_search_set_terminal (self->tsearch, VTE_TERMINAL (self->terminal_top));
+  ide_terminal_search_set_terminal (self->tsearch, VTE_TERMINAL (self->terminal));
 
   ide_terminal_page_actions_init (self);
 
-  style_context = gtk_widget_get_style_context (GTK_WIDGET (self->terminal_top));
-  gtk_style_context_add_class (style_context, "terminal");
-  g_signal_connect_object (style_context,
-                           "changed",
-                           G_CALLBACK (style_context_changed),
-                           self,
-                           0);
-  style_context_changed (style_context, self);
-
-  gtk_widget_set_can_focus (GTK_WIDGET (self->terminal_top), TRUE);
-
   ide_widget_set_context_handler (self, ide_terminal_page_context_set);
 }
 
+/**
+ * ide_terminal_page_get_pty:
+ * @self: a #IdeTerminalPage
+ *
+ * Gets the #VtePty for the page.
+ *
+ * Returns: (transfer none): a #VtePty
+ */
+VtePty *
+ide_terminal_page_get_pty (IdeTerminalPage *self)
+{
+  g_return_val_if_fail (IDE_IS_TERMINAL_PAGE (self), NULL);
+
+  return self->pty;
+}
+
 void
 ide_terminal_page_set_pty (IdeTerminalPage *self,
                            VtePty          *pty)
@@ -636,10 +572,13 @@ ide_terminal_page_set_pty (IdeTerminalPage *self,
   g_return_if_fail (IDE_IS_TERMINAL_PAGE (self));
   g_return_if_fail (VTE_IS_PTY (pty));
 
+  if (self->destroyed)
+    return;
+
   if (g_set_object (&self->pty, pty))
     {
-      vte_terminal_reset (VTE_TERMINAL (self->terminal_top), TRUE, TRUE);
-      vte_terminal_set_pty (VTE_TERMINAL (self->terminal_top), pty);
+      vte_terminal_reset (VTE_TERMINAL (self->terminal), TRUE, TRUE);
+      vte_terminal_set_pty (VTE_TERMINAL (self->terminal), pty);
     }
 }
 
@@ -648,9 +587,10 @@ ide_terminal_page_feed (IdeTerminalPage *self,
                         const gchar     *message)
 {
   g_return_if_fail (IDE_IS_TERMINAL_PAGE (self));
+  g_return_if_fail (self->destroyed == FALSE);
 
-  if (self->terminal_top != NULL)
-    vte_terminal_feed (VTE_TERMINAL (self->terminal_top), message, -1);
+  if (self->terminal != NULL)
+    vte_terminal_feed (VTE_TERMINAL (self->terminal), message, -1);
 }
 
 void
@@ -659,24 +599,12 @@ ide_terminal_page_set_launcher (IdeTerminalPage     *self,
 {
   g_return_if_fail (IDE_IS_TERMINAL_PAGE (self));
   g_return_if_fail (!launcher || IDE_IS_TERMINAL_LAUNCHER (launcher));
+  g_return_if_fail (self->destroyed == FALSE);
 
   if (g_set_object (&self->launcher, launcher))
     {
-      gboolean can_split;
-
-      if (launcher != NULL)
-        {
-          const gchar *title = ide_terminal_launcher_get_title (launcher);
-          ide_page_set_title (IDE_PAGE (self), title);
-          can_split = ide_terminal_launcher_can_respawn (launcher);
-        }
-      else
-        {
-          self->manage_spawn = FALSE;
-          can_split = FALSE;
-        }
-
-      ide_page_set_can_split (IDE_PAGE (self), can_split);
+      ide_page_set_can_split (IDE_PAGE (self), TRUE);
+      g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_LAUNCHER]);
     }
 }
 
@@ -684,6 +612,89 @@ const gchar *
 ide_terminal_page_get_current_directory_uri (IdeTerminalPage *self)
 {
   g_return_val_if_fail (IDE_IS_TERMINAL_PAGE (self), NULL);
+  g_return_val_if_fail (self->destroyed == FALSE, NULL);
 
-  return vte_terminal_get_current_directory_uri (VTE_TERMINAL (self->terminal_top));
+  return vte_terminal_get_current_directory_uri (VTE_TERMINAL (self->terminal));
+}
+
+/**
+ * ide_terminal_page_get_launcher:
+ * @self: a #IdeTerminalPage
+ *
+ * Gets the launcher for the page.
+ *
+ * Returns: (transfer none) (nullable): an #IdeTerminalLauncher or %NULL
+ */
+IdeTerminalLauncher *
+ide_terminal_page_get_launcher (IdeTerminalPage *self)
+{
+  g_return_val_if_fail (IDE_IS_TERMINAL_PAGE (self), NULL);
+
+  return self->launcher;
+}
+
+gboolean
+ide_terminal_page_has_exited (IdeTerminalPage *self)
+{
+  g_return_val_if_fail (IDE_IS_TERMINAL_PAGE (self), FALSE);
+
+  return self->exited;
+}
+
+/**
+ * ide_terminal_page_get_terminal:
+ * @self: a #IdeTerminalPage
+ *
+ * Gets the underlying terminal widget.
+ *
+ * Returns: (transfer none): an #IdeTerminal
+ */
+IdeTerminal *
+ide_terminal_page_get_terminal (IdeTerminalPage *self)
+{
+  g_return_val_if_fail (IDE_IS_TERMINAL_PAGE (self), NULL);
+
+  return self->terminal;
+}
+
+IdeTerminalPage *
+ide_terminal_page_new_completed (const char *title,
+                                 const char *text,
+                                 int         columns,
+                                 int         rows)
+{
+  g_autofree char *copy = g_strdup (text);
+  IdeTerminalPage *self;
+  IdeLineReader reader;
+  char *line;
+  gsize line_len;
+
+  self = g_object_new (IDE_TYPE_TERMINAL_PAGE,
+                       "title", title,
+                       NULL);
+
+  vte_terminal_set_input_enabled (VTE_TERMINAL (self->terminal), FALSE);
+
+  self->close_on_exit = FALSE;
+  self->manage_spawn = FALSE;
+  g_clear_object (&self->launcher);
+  self->respawn_on_exit = FALSE;
+  self->exited = TRUE;
+
+  if (columns > 0 && rows > 0)
+    vte_terminal_set_size (VTE_TERMINAL (self->terminal), columns, rows);
+
+  ide_line_reader_init (&reader, copy, -1);
+  while ((line = ide_line_reader_next (&reader, &line_len)))
+    {
+      gboolean had_newline = line[line_len] != 0;
+
+      line[line_len] = 0;
+      ide_terminal_page_feed (self, line);
+
+      if (had_newline)
+        ide_terminal_page_feed (self, "\r\n");
+    }
+
+  return self;
 }
