@@ -22,20 +22,23 @@
 
 #include "config.h"
 
-#include <dazzle.h>
 #include <glib/gi18n.h>
+
 #include <libide-gui.h>
 
 #include "ide-terminal.h"
+#include "ide-terminal-search.h"
 
 #define BUILDER_PCRE2_MULTILINE 0x00000400u
 #define BUILDER_PCRE2_UCP 0x00020000u
 
 typedef struct
 {
-  GtkWidget *popup_menu;
-  GSettings *settings;
-  gchar     *url;
+  GtkPopover *popover;
+  GSettings  *settings;
+  char       *url;
+  GdkRGBA     bg;
+  GdkRGBA     fg;
 } IdeTerminalPrivate;
 
 typedef struct
@@ -52,21 +55,11 @@ typedef struct
 
 G_DEFINE_TYPE_WITH_PRIVATE (IdeTerminal, ide_terminal, VTE_TYPE_TERMINAL)
 
-enum {
-  COPY_LINK_ADDRESS,
-  OPEN_LINK,
-  POPULATE_POPUP,
-  SELECT_ALL,
-  SEARCH_REVEAL,
-  N_SIGNALS
-};
-
 /* From vteapp.c */
 #define DINGUS1 "(((gopher|news|telnet|nntp|file|http|ftp|https)://)|(www|ftp)[-A-Za-z0-9]*\\.)[-A-Za-z0-9\\.]+(:[0-9]*)?"
 #define DINGUS2 DINGUS1 "/[-A-Za-z0-9_\\$\\.\\+\\!\\*\\(\\),;:@&=\\?/~\\#\\%]*[^]'\\.}>\\) ,\\\"]"
 #define FILENAME_PLUS_LOCATION "(?<filename>[[:alnum:]\\+\\-\\.\\/_]+):(?<line>\\d+):(?<column>\\d+)"
 
-static guint signals[N_SIGNALS];
 static const gchar *url_regexes[] = { DINGUS1, DINGUS2, FILENAME_PLUS_LOCATION };
 static GRegex *filename_regex;
 static const GdkRGBA solarized_palette[] = {
@@ -92,204 +85,197 @@ static const GdkRGBA solarized_palette[] = {
   { 0.992156, 0.964705, 0.890196, 1 },
 };
 
-static void
-style_context_changed (IdeTerminal *self,
-                       GtkStyleContext *style_context)
+void
+ide_terminal_get_colors (IdeTerminal *self,
+                         GdkRGBA     *bg,
+                         GdkRGBA     *fg)
 {
-  GtkStateFlags state;
+  IdeTerminalPrivate *priv = ide_terminal_get_instance_private (self);
+
+  g_return_if_fail (IDE_IS_TERMINAL (self));
+
+  if (bg)
+    *bg = priv->bg;
+
+  if (fg)
+    *fg = priv->fg;
+}
+
+static void
+ide_terminal_css_changed (GtkWidget         *widget,
+                          GtkCssStyleChange *change)
+{
+  IdeTerminal *self = (IdeTerminal *)widget;
+  IdeTerminalPrivate *priv = ide_terminal_get_instance_private (self);
+  GtkStyleContext *style_context;
   GdkRGBA fg;
   GdkRGBA bg;
 
-  g_assert (GTK_IS_STYLE_CONTEXT (style_context));
-  g_assert (IDE_IS_TERMINAL (self));
+  g_assert (IDE_IS_TERMINAL (widget));
 
-  state = gtk_style_context_get_state (style_context);
+  style_context = gtk_widget_get_style_context (widget);
 
-  G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
-  gtk_style_context_get_color (style_context, state, &fg);
-  gtk_style_context_get_background_color (style_context, state, &bg);
-  G_GNUC_END_IGNORE_DEPRECATIONS;
+  if (!gtk_style_context_lookup_color (style_context, "window_fg_color", &fg))
+    gdk_rgba_parse (&fg, "#eeeeec");
 
-  if (bg.alpha == 0.0)
-    gdk_rgba_parse (&bg, "#f6f7f8");
+  if (!gtk_style_context_lookup_color (style_context, "window_bg_color", &bg))
+    gdk_rgba_parse (&bg, "#242424");
 
-  vte_terminal_set_colors (VTE_TERMINAL (self), &fg, &bg,
+  vte_terminal_set_colors (VTE_TERMINAL (widget),
+                           &fg, &bg,
                            solarized_palette, G_N_ELEMENTS (solarized_palette));
+
+  priv->fg = fg;
+  priv->bg = bg;
 }
 
 static void
-popup_menu_detach (GtkWidget *attach_widget,
-                   GtkMenu   *menu)
+ide_terminal_update_clipboard_actions (IdeTerminal *self)
+{
+  GdkClipboard *clipboard;
+  gboolean can_paste;
+  gboolean has_selection;
+
+  g_assert (IDE_IS_TERMINAL (self));
+
+  clipboard = gtk_widget_get_clipboard (GTK_WIDGET (self));
+  can_paste = gdk_content_formats_contain_gtype (gdk_clipboard_get_formats (clipboard), G_TYPE_STRING);
+  has_selection = vte_terminal_get_has_selection (VTE_TERMINAL (self));
+
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "clipboard.copy", has_selection);
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "clipboard.paste", can_paste);
+}
+
+static void
+ide_terminal_update_url_actions (IdeTerminal *self,
+                                 double       x,
+                                 double       y)
+{
+  IdeTerminalPrivate *priv = ide_terminal_get_instance_private (self);
+  g_autofree char *pattern = NULL;
+  int tag = 0;
+
+  g_assert (IDE_IS_TERMINAL (self));
+
+  pattern = vte_terminal_check_match_at (VTE_TERMINAL (self), x, y, &tag);
+
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "clipboard.copy-link", pattern != NULL);
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "terminal.open-link", pattern != NULL);
+
+  ide_set_string (&priv->url, pattern);
+}
+
+static gboolean
+clear_url_actions_cb (gpointer data)
+{
+  IdeTerminal *self = data;
+
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "clipboard.copy-link", FALSE);
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "terminal.open-link", FALSE);
+
+  return G_SOURCE_REMOVE;
+}
+
+static void
+ide_terminal_popover_closed_cb (IdeTerminal *self,
+                                GtkPopover  *popover)
+{
+  g_assert (IDE_IS_TERMINAL (self));
+  g_assert (GTK_IS_POPOVER (popover));
+
+  g_idle_add_full (G_PRIORITY_LOW,
+                   clear_url_actions_cb,
+                   g_object_ref (self),
+                   g_object_unref);
+}
+
+static void
+ide_terminal_popup (IdeTerminal *self,
+                    double       x,
+                    double       y)
 {
-  IdeTerminal *self = IDE_TERMINAL (attach_widget);
   IdeTerminalPrivate *priv = ide_terminal_get_instance_private (self);
 
   g_assert (IDE_IS_TERMINAL (self));
-  g_assert (GTK_IS_MENU (menu));
-  g_assert (priv->popup_menu == NULL || GTK_IS_WIDGET (priv->popup_menu));
 
-  g_clear_pointer (&priv->url, g_free);
+  ide_terminal_update_clipboard_actions (self);
+  ide_terminal_update_url_actions (self, x, y);
 
-  if (priv->popup_menu == GTK_WIDGET (menu))
-    g_clear_pointer (&priv->popup_menu, gtk_widget_destroy);
-}
-
-static void
-popup_targets_received (GtkClipboard     *clipboard,
-                        GtkSelectionData *data,
-                        gpointer          user_data)
-{
-  PopupInfo *popup_info = user_data;
-  g_autoptr(IdeTerminal) self = NULL;
-  g_autoptr(GdkEvent) event = NULL;
-  IdeTerminalPrivate *priv;
-
-  g_assert (popup_info != NULL);
-  g_assert (IDE_IS_TERMINAL (popup_info->terminal));
-
-  self = g_steal_pointer (&popup_info->terminal);
-  priv = ide_terminal_get_instance_private (self);
-  event = g_steal_pointer (&popup_info->event);
-
-  if (gtk_widget_get_realized (GTK_WIDGET (self)))
+  if (priv->popover == NULL)
     {
-      DzlWidgetActionGroup *group;
-      GMenu *menu;
-      gboolean clipboard_contains_text;
-      gboolean have_selection;
+      GMenu *menu = ide_application_get_menu_by_id (IDE_APPLICATION_DEFAULT, "ide-terminal-page-popup-menu");
 
-      clipboard_contains_text = gtk_selection_data_targets_include_text (data);
-      have_selection = vte_terminal_get_has_selection (VTE_TERMINAL (self));
+      priv->popover = GTK_POPOVER (gtk_popover_menu_new_from_model (G_MENU_MODEL (menu)));
+      gtk_popover_set_has_arrow (priv->popover, FALSE);
+      gtk_widget_set_halign (GTK_WIDGET (priv->popover), GTK_ALIGN_END);
+      gtk_widget_set_parent (GTK_WIDGET (priv->popover), GTK_WIDGET (self));
 
-      g_clear_pointer (&priv->popup_menu, gtk_widget_destroy);
-
-      priv->url = vte_terminal_match_check_event (VTE_TERMINAL (self), event, NULL);
-
-      menu = dzl_application_get_menu_by_id (DZL_APPLICATION_DEFAULT, "ide-terminal-page-popup-menu");
-      priv->popup_menu = gtk_menu_new_from_model (G_MENU_MODEL (menu));
-
-      group = DZL_WIDGET_ACTION_GROUP (gtk_widget_get_action_group (GTK_WIDGET (self), "terminal"));
-
-      dzl_widget_action_group_set_action_enabled (group, "copy-link-address", priv->url != NULL);
-      dzl_widget_action_group_set_action_enabled (group, "open-link", priv->url != NULL);
-      dzl_widget_action_group_set_action_enabled (group, "copy-clipboard", have_selection);
-      dzl_widget_action_group_set_action_enabled (group, "paste-clipboard", clipboard_contains_text);
-
-      dzl_gtk_widget_add_style_class (priv->popup_menu, GTK_STYLE_CLASS_CONTEXT_MENU);
-      gtk_menu_attach_to_widget (GTK_MENU (priv->popup_menu), GTK_WIDGET (self), popup_menu_detach);
-
-      g_signal_emit (self, signals[POPULATE_POPUP], 0, priv->popup_menu);
-
-      gtk_menu_popup_at_pointer (GTK_MENU (priv->popup_menu), event);
+      g_signal_connect_object (priv->popover,
+                               "closed",
+                               G_CALLBACK (ide_terminal_popover_closed_cb),
+                               self,
+                               G_CONNECT_SWAPPED);
     }
 
-  g_slice_free (PopupInfo, popup_info);
+  gtk_popover_set_pointing_to (priv->popover,
+                               &(GdkRectangle) { x, y, 1, 1 });
+
+  gtk_popover_popup (priv->popover);
 }
 
 static void
-ide_terminal_do_popup (IdeTerminal    *self,
-                       const GdkEvent *event)
+ide_terminal_click_pressed_cb (IdeTerminal     *self,
+                               int              n_presses,
+                               double           x,
+                               double           y,
+                               GtkGestureClick *click)
 {
-  PopupInfo *popup_info;
-  GtkClipboard *clipboard;
-
-  g_assert (IDE_IS_TERMINAL (self));
-
-  popup_info = g_slice_new0 (PopupInfo);
-  popup_info->event = event ? gdk_event_copy (event) : gtk_get_current_event ();
-  popup_info->terminal = g_object_ref (self);
-
-  clipboard = gtk_widget_get_clipboard (GTK_WIDGET (self), GDK_SELECTION_CLIPBOARD);
-
-  gtk_clipboard_request_contents (clipboard,
-                                  gdk_atom_intern_static_string ("TARGETS"),
-                                  popup_targets_received,
-                                  popup_info);
-}
-
-static gboolean
-ide_terminal_popup_menu (GtkWidget *widget)
-{
-  IdeTerminal *self = (IdeTerminal *)widget;
-
-  g_assert (IDE_IS_TERMINAL (self));
-
-  ide_terminal_do_popup (self, NULL);
-
-  return TRUE;
-}
-
-static gboolean
-ide_terminal_button_press_event (GtkWidget      *widget,
-                                 GdkEventButton *button)
-{
-  IdeTerminal *self = (IdeTerminal *)widget;
   IdeTerminalPrivate *priv = ide_terminal_get_instance_private (self);
+  int button;
+
+  IDE_ENTRY;
 
   g_assert (IDE_IS_TERMINAL (self));
-  g_assert (button != NULL);
+  g_assert (GTK_IS_GESTURE_CLICK (click));
 
-  if (button->type == GDK_BUTTON_PRESS)
+  ide_terminal_update_url_actions (self, x, y);
+
+  button = gtk_gesture_single_get_current_button (GTK_GESTURE_SINGLE (click));
+
+  if (button == 1)
     {
-      if (button->button == GDK_BUTTON_PRIMARY)
+      if (priv->url != NULL)
         {
-          g_autofree gchar *pattern = NULL;
-
-          pattern = vte_terminal_match_check_event (VTE_TERMINAL (self), (GdkEvent *)button, NULL);
-
-          if (pattern != NULL)
-            {
-              gboolean ret = GDK_EVENT_PROPAGATE;
-
-              g_free (priv->url);
-              priv->url = g_steal_pointer (&pattern);
-
-              g_signal_emit (self, signals [OPEN_LINK], 0, &ret);
-
-              return ret;
-            }
-        }
-      else if (button->button == GDK_BUTTON_SECONDARY)
-        {
-          if (!gtk_widget_has_focus (GTK_WIDGET (self)))
-            gtk_widget_grab_focus (GTK_WIDGET (self));
-
-          ide_terminal_do_popup (self, (GdkEvent *)button);
-
-          return GDK_EVENT_STOP;
+          gtk_widget_activate_action (GTK_WIDGET (self), "terminal.open-link", NULL);
+          gtk_gesture_set_state (GTK_GESTURE (click), GTK_EVENT_SEQUENCE_CLAIMED);
         }
     }
+  else if (button == 3)
+    {
+      if (!gtk_widget_has_focus (GTK_WIDGET (self)))
+        gtk_widget_grab_focus (GTK_WIDGET (self));
 
-  return GTK_WIDGET_CLASS (ide_terminal_parent_class)->button_press_event (widget, button);
+      ide_terminal_popup (self, x, y);
+
+      gtk_gesture_set_state (GTK_GESTURE (click), GTK_EVENT_SEQUENCE_CLAIMED);
+    }
+
+  IDE_EXIT;
 }
 
 static void
-ide_terminal_real_select_all (IdeTerminal *self,
-                              gboolean     all)
-{
-  g_assert (IDE_IS_TERMINAL (self));
-
-  if (all)
-    vte_terminal_select_all (VTE_TERMINAL (self));
-  else
-    vte_terminal_unselect_all (VTE_TERMINAL (self));
-}
-
-static gboolean
-ide_terminal_copy_link_address (IdeTerminal *self)
+copy_link_address_action (GtkWidget  *widget,
+                          const char *action_name,
+                          GVariant   *param)
 {
+  IdeTerminal *self = (IdeTerminal *)widget;
   IdeTerminalPrivate *priv = ide_terminal_get_instance_private (self);
 
   g_assert (IDE_IS_TERMINAL (self));
   g_assert (priv->url != NULL);
 
-  if (ide_str_empty0 (priv->url))
-    return FALSE;
-
-  gtk_clipboard_set_text (gtk_widget_get_clipboard (GTK_WIDGET (self), GDK_SELECTION_CLIPBOARD),
-                          priv->url, strlen (priv->url));
-
-  return TRUE;
+  if (!ide_str_empty0 (priv->url))
+    gdk_clipboard_set_text (gtk_widget_get_clipboard (GTK_WIDGET (self)), priv->url);
 }
 
 static void
@@ -312,27 +298,29 @@ ide_terminal_open_link_resolve_cb (GObject      *object,
                                  pos->line,
                                  pos->column,
                                  IDE_BUFFER_OPEN_FLAGS_NONE,
-                                 NULL, NULL, NULL);
+                                 NULL, NULL, NULL, NULL);
 
   g_slice_free (Position, pos);
 }
 
-static gboolean
-ide_terminal_open_link (IdeTerminal *self)
+static void
+open_link_action (GtkWidget  *widget,
+                  const char *action_name,
+                  GVariant   *param)
 {
+  IdeTerminal *self = (IdeTerminal *)widget;
   IdeTerminalPrivate *priv = ide_terminal_get_instance_private (self);
   g_autoptr(GMatchInfo) match = NULL;
-  g_autofree gchar *filename = NULL;
-  g_autofree gchar *line = NULL;
-  g_autofree gchar *column = NULL;
+  g_autofree char *filename = NULL;
+  g_autofree char *line = NULL;
+  g_autofree char *column = NULL;
   GtkApplication *app;
   GtkWindow *focused_window;
 
   g_assert (IDE_IS_TERMINAL (self));
-  g_assert (priv->url != NULL);
 
   if (ide_str_empty0 (priv->url))
-    return FALSE;
+    return;
 
   if (g_regex_match (filename_regex, priv->url, 0, &match) &&
       (filename = g_match_info_fetch (match, 1)) &&
@@ -353,22 +341,38 @@ ide_terminal_open_link (IdeTerminal *self)
                                         ide_terminal_open_link_resolve_cb,
                                         g_slice_dup (Position, &pos));
 
-      return TRUE;
+      return;
     }
 
   if ((app = GTK_APPLICATION (g_application_get_default ())) &&
       (focused_window = gtk_application_get_active_window (app)))
-    return ide_gtk_show_uri_on_window (focused_window,
-                                       priv->url,
-                                       g_get_monotonic_time (),
-                                       NULL);
+    ide_gtk_show_uri_on_window (focused_window,
+                                priv->url,
+                                g_get_monotonic_time (),
+                                NULL);
+}
 
-  return FALSE;
+static GtkWidget *
+find_child_typed (GtkWidget *parent,
+                  GType      child_type)
+{
+  for (GtkWidget *child = gtk_widget_get_first_child (parent);
+       child;
+       child = gtk_widget_get_next_sibling (child))
+    {
+      if (g_type_is_a (G_OBJECT_TYPE (child), child_type))
+        return child;
+    }
+
+  return NULL;
 }
 
 static void
-ide_terminal_real_search_reveal (IdeTerminal *self)
+ide_terminal_search_reveal (GtkWidget  *widget,
+                            const char *action_name,
+                            GVariant   *param)
 {
+  IdeTerminal *self = (IdeTerminal *)widget;
   GtkWidget *parent_overlay;
 
   g_assert (IDE_IS_TERMINAL (self));
@@ -377,10 +381,15 @@ ide_terminal_real_search_reveal (IdeTerminal *self)
 
   if (parent_overlay != NULL)
     {
-      GtkRevealer *revealer = dzl_gtk_widget_find_child_typed (parent_overlay, GTK_TYPE_REVEALER);
+      IdeTerminalSearch *search = IDE_TERMINAL_SEARCH (find_child_typed (parent_overlay, IDE_TYPE_TERMINAL_SEARCH));
 
-      if (revealer != NULL && !gtk_revealer_get_child_revealed (revealer))
-        gtk_revealer_set_reveal_child (revealer, TRUE);
+      if (search != NULL)
+        {
+          GtkRevealer *revealer = ide_terminal_search_get_revealer (search);
+
+          if (!gtk_revealer_get_child_revealed (revealer))
+            gtk_revealer_set_reveal_child (revealer, TRUE);
+        }
     }
 }
 
@@ -405,33 +414,32 @@ ide_terminal_font_changed (IdeTerminal *self,
 }
 
 static void
-ide_terminal_size_allocate (GtkWidget     *widget,
-                            GtkAllocation *alloc)
+ide_terminal_size_allocate (GtkWidget *widget,
+                            int        width,
+                            int        height,
+                            int        baseline)
 {
   IdeTerminal *self = (IdeTerminal *)widget;
-  glong width;
-  glong height;
-  glong columns;
-  glong rows;
+  IdeTerminalPrivate *priv = ide_terminal_get_instance_private (self);
+  int char_width, char_height;
+  int columns, rows;
 
-  GTK_WIDGET_CLASS (ide_terminal_parent_class)->size_allocate (widget, alloc);
+  GTK_WIDGET_CLASS (ide_terminal_parent_class)->size_allocate (widget, width, height, baseline);
 
-  if ((alloc->width == 0) || (alloc->height == 0))
+  char_width = vte_terminal_get_char_width (VTE_TERMINAL (self));
+  char_height = vte_terminal_get_char_height (VTE_TERMINAL (self));
+  if (char_width == 0 || char_height == 0)
     return;
 
-  width = vte_terminal_get_char_width (VTE_TERMINAL (self));
-  height = vte_terminal_get_char_height (VTE_TERMINAL (self));
-
-  if ((width == 0) || (height == 0))
-    return;
-
-  columns = alloc->width / width;
-  rows = alloc->height / height;
-
-  if ((columns < 2) || (rows < 2))
+  columns = width / char_width;
+  rows = height / char_height;
+  if (columns < 2 || rows < 2)
     return;
 
   vte_terminal_set_size (VTE_TERMINAL (self), columns, rows);
+
+  if (priv->popover)
+    gtk_popover_present (priv->popover);
 }
 
 static void
@@ -455,113 +463,122 @@ update_scrollback_cb (IdeTerminal *self,
 }
 
 static void
-ide_terminal_destroy (GtkWidget *widget)
+select_all_action (GtkWidget  *widget,
+                   const char *action_name,
+                   GVariant   *param)
 {
-  IdeTerminal *self = (IdeTerminal *)widget;
+  g_assert (IDE_IS_TERMINAL (widget));
+  g_assert (g_variant_is_of_type (param, G_VARIANT_TYPE_BOOLEAN));
+
+  if (g_variant_get_boolean (param))
+    vte_terminal_select_all (VTE_TERMINAL (widget));
+  else
+    vte_terminal_unselect_all (VTE_TERMINAL (widget));
+}
+
+static void
+copy_clipboard_action (GtkWidget  *widget,
+                       const char *action_name,
+                       GVariant   *param)
+{
+  GdkClipboard *clipboard = gtk_widget_get_clipboard (widget);
+  g_autofree char *text = vte_terminal_get_text_selected (VTE_TERMINAL (widget), VTE_FORMAT_TEXT);
+  gdk_clipboard_set_text (clipboard, text);
+}
+
+static void
+ide_terminal_read_text_cb (GObject      *object,
+                           GAsyncResult *result,
+                           gpointer      user_data)
+{
+  GdkClipboard *clipboard = (GdkClipboard *)object;
+  g_autoptr(IdeTerminal) self = user_data;
+  g_autofree char *text = NULL;
+
+  g_assert (GDK_IS_CLIPBOARD (clipboard));
+  g_assert (G_IS_ASYNC_RESULT (result));
+  g_assert (IDE_IS_TERMINAL (self));
+
+  if ((text = gdk_clipboard_read_text_finish (clipboard, result, NULL)))
+    vte_terminal_paste_text (VTE_TERMINAL (self), text);
+}
+
+static void
+paste_clipboard_action (GtkWidget  *widget,
+                        const char *action_name,
+                        GVariant   *param)
+{
+  GdkClipboard *clipboard = gtk_widget_get_clipboard (widget);
+
+  gdk_clipboard_read_text_async (clipboard,
+                                 NULL,
+                                 ide_terminal_read_text_cb,
+                                 g_object_ref (widget));
+}
+
+static void
+ide_terminal_selection_changed (VteTerminal *terminal)
+{
+  ide_terminal_update_clipboard_actions (IDE_TERMINAL (terminal));
+}
+
+static void
+ide_terminal_dispose (GObject *object)
+{
+  IdeTerminal *self = (IdeTerminal *)object;
   IdeTerminalPrivate *priv = ide_terminal_get_instance_private (self);
 
-  g_assert (IDE_IS_TERMINAL (self));
+  g_clear_pointer ((GtkWidget **)&priv->popover, gtk_widget_unparent);
 
   g_clear_object (&priv->settings);
   g_clear_pointer (&priv->url, g_free);
 
-  GTK_WIDGET_CLASS (ide_terminal_parent_class)->destroy (widget);
+  G_OBJECT_CLASS (ide_terminal_parent_class)->dispose (object);
 }
 
 static void
 ide_terminal_class_init (IdeTerminalClass *klass)
 {
+  GObjectClass *object_class = G_OBJECT_CLASS (klass);
   GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
-  GtkBindingSet *binding_set;
+  VteTerminalClass *terminal_class = VTE_TERMINAL_CLASS (klass);
 
-  widget_class->destroy = ide_terminal_destroy;
-  widget_class->button_press_event = ide_terminal_button_press_event;
-  widget_class->popup_menu = ide_terminal_popup_menu;
+  object_class->dispose = ide_terminal_dispose;
+
+  widget_class->css_changed = ide_terminal_css_changed;
   widget_class->size_allocate = ide_terminal_size_allocate;
 
-  klass->copy_link_address = ide_terminal_copy_link_address;
-  klass->open_link = ide_terminal_open_link;
-  klass->select_all = ide_terminal_real_select_all;
-  klass->search_reveal = ide_terminal_real_search_reveal;
+  terminal_class->selection_changed = ide_terminal_selection_changed;
 
   filename_regex = g_regex_new (FILENAME_PLUS_LOCATION, 0, 0, NULL);
   g_assert (filename_regex != NULL);
 
-  signals [COPY_LINK_ADDRESS] =
-    g_signal_new ("copy-link-address",
-                  G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
-                  G_STRUCT_OFFSET (IdeTerminalClass, copy_link_address),
-                  NULL, NULL, NULL,
-                  G_TYPE_BOOLEAN,
-                  0);
+  gtk_widget_class_install_action (widget_class, "clipboard.copy", NULL, copy_clipboard_action);
+  gtk_widget_class_install_action (widget_class, "clipboard.copy-link", NULL, copy_link_address_action);
+  gtk_widget_class_install_action (widget_class, "clipboard.paste", NULL, paste_clipboard_action);
+  gtk_widget_class_install_action (widget_class, "terminal.open-link", NULL, open_link_action);
+  gtk_widget_class_install_action (widget_class, "terminal.search", NULL, ide_terminal_search_reveal);
+  gtk_widget_class_install_action (widget_class, "terminal.select-all", "b", select_all_action);
 
-  signals [SEARCH_REVEAL] =
-    g_signal_new ("search-reveal",
-                  G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
-                  G_STRUCT_OFFSET (IdeTerminalClass, search_reveal),
-                  NULL, NULL, NULL,
-                  G_TYPE_NONE,
-                  0);
-
-  signals [OPEN_LINK] =
-    g_signal_new ("open-link",
-                  G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
-                  G_STRUCT_OFFSET (IdeTerminalClass, open_link),
-                  NULL, NULL, NULL,
-                  G_TYPE_BOOLEAN,
-                  0);
-
-  signals [POPULATE_POPUP] =
-    g_signal_new ("populate-popup",
-                  G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST,
-                  G_STRUCT_OFFSET (IdeTerminalClass, populate_popup),
-                  NULL, NULL, NULL,
-                  G_TYPE_NONE,
-                  1,
-                  GTK_TYPE_WIDGET);
-
-  signals [SELECT_ALL] =
-    g_signal_new ("select-all",
-                  G_TYPE_FROM_CLASS (klass),
-                  G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
-                  G_STRUCT_OFFSET (IdeTerminalClass, select_all),
-                  NULL, NULL, NULL,
-                  G_TYPE_NONE,
-                  1,
-                  G_TYPE_BOOLEAN);
-
-  binding_set = gtk_binding_set_by_class (klass);
-
-  gtk_binding_entry_add_signal (binding_set,
-                                GDK_KEY_c,
-                                GDK_SHIFT_MASK | GDK_CONTROL_MASK,
-                                "copy-clipboard",
-                                0);
-
-  gtk_binding_entry_add_signal (binding_set,
-                                GDK_KEY_v,
-                                GDK_SHIFT_MASK | GDK_CONTROL_MASK,
-                                "paste-clipboard",
-                                0);
-
-  gtk_binding_entry_add_signal (binding_set,
-                                GDK_KEY_f,
-                                GDK_SHIFT_MASK | GDK_CONTROL_MASK,
-                                "search-reveal",
-                                0);
+  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_c, GDK_CONTROL_MASK|GDK_SHIFT_MASK, "clipboard.copy", NULL);
+  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_v, GDK_CONTROL_MASK|GDK_SHIFT_MASK, "clipboard.paste", NULL);
 }
 
 static void
 ide_terminal_init (IdeTerminal *self)
 {
   IdeTerminalPrivate *priv = ide_terminal_get_instance_private (self);
-  GtkStyleContext *style_context;
+  GtkEventController *gesture;
 
-  dzl_widget_action_group_attach (self, "terminal");
+  gesture = GTK_EVENT_CONTROLLER (gtk_gesture_click_new ());
+  gtk_gesture_single_set_button (GTK_GESTURE_SINGLE (gesture), 0);
+  gtk_event_controller_set_propagation_phase (gesture, GTK_PHASE_CAPTURE);
+  g_signal_connect_object (gesture,
+                           "pressed",
+                           G_CALLBACK (ide_terminal_click_pressed_cb),
+                           self,
+                           G_CONNECT_SWAPPED);
+  gtk_widget_add_controller (GTK_WIDGET (self), g_steal_pointer (&gesture));
 
   for (guint i = 0; i < G_N_ELEMENTS (url_regexes); i++)
     {
@@ -569,7 +586,7 @@ ide_terminal_init (IdeTerminal *self)
       const gchar *pattern = url_regexes[i];
       gint tag;
 
-      regex = vte_regex_new_for_match (pattern, DZL_LITERAL_LENGTH (pattern),
+      regex = vte_regex_new_for_match (pattern, strlen (pattern),
                                        VTE_REGEX_FLAGS_DEFAULT | BUILDER_PCRE2_MULTILINE | BUILDER_PCRE2_UCP,
                                        NULL);
       tag = vte_terminal_match_add_regex (VTE_TERMINAL (self), regex, 0);
@@ -599,17 +616,6 @@ ide_terminal_init (IdeTerminal *self)
   ide_terminal_font_changed (self, NULL, priv->settings);
   update_scrollback_cb (self, "scrollback-lines", priv->settings);
 
-  style_context = gtk_widget_get_style_context (GTK_WIDGET (self));
-  gtk_style_context_add_class (style_context, "terminal");
-  g_signal_connect_object (style_context,
-                           "changed",
-                           G_CALLBACK (style_context_changed),
-                           self,
-                           G_CONNECT_SWAPPED);
-  style_context_changed (self, style_context);
-
-  gtk_widget_set_can_focus (GTK_WIDGET (self), TRUE);
-
   vte_terminal_set_enable_fallback_scrolling (VTE_TERMINAL (self), FALSE);
   vte_terminal_set_scroll_unit_is_pixels (VTE_TERMINAL (self), TRUE);
 }
