@@ -62,7 +62,7 @@ populate_free (Populate *state)
 static GType
 ide_ctags_results_get_item_type (GListModel *model)
 {
-  return IDE_TYPE_COMPLETION_PROPOSAL;
+  return GTK_SOURCE_TYPE_COMPLETION_PROPOSAL;
 }
 
 static gpointer
@@ -73,7 +73,9 @@ ide_ctags_results_get_item (GListModel *model,
   const Item *item;
 
   g_assert (IDE_IS_CTAGS_RESULTS (self));
-  g_assert (position < self->items->len);
+
+  if (position >= self->items->len)
+    return NULL;
 
   item = &g_array_index (self->items, Item, position);
 
@@ -187,7 +189,7 @@ ide_ctags_results_populate_worker (IdeTask      *task,
           if (!ide_ctags_is_allowed (entry, p->suffixes))
             continue;
 
-          if (ide_completion_fuzzy_match (entry->name, p->casefold, &priority))
+          if (gtk_source_completion_fuzzy_match (entry->name, p->casefold, &priority))
             {
               Item item;
 
