@@ -22,6 +22,7 @@
 
 #include "config.h"
 
+#include <libide-gui.h>
 #include <libide-projects.h>
 #include <libide-tree.h>
 
@@ -30,9 +31,9 @@
 
 struct _GbpGrepTreeAddin
 {
-  GObject  parent_instance;
-
-  IdeTree *tree;
+  GObject             parent_instance;
+  IdeTree            *tree;
+  GSimpleActionGroup *group;
 };
 
 static void
@@ -72,12 +73,13 @@ gbp_grep_tree_addin_load (IdeTreeAddin *addin,
                           IdeTree      *tree,
                           IdeTreeModel *model)
 {
-  GbpGrepTreeAddin *self = (GbpGrepTreeAddin *)addin;
-  g_autoptr(GActionMap) group = NULL;
   static const GActionEntry actions[] = {
     { "find-in-files", find_in_files_action },
   };
 
+  GbpGrepTreeAddin *self = (GbpGrepTreeAddin *)addin;
+  GtkWidget *pane;
+
   g_assert (IDE_IS_MAIN_THREAD ());
   g_assert (GBP_IS_GREP_TREE_ADDIN (self));
   g_assert (IDE_IS_TREE (tree));
@@ -85,9 +87,17 @@ gbp_grep_tree_addin_load (IdeTreeAddin *addin,
 
   self->tree = tree;
 
-  group = G_ACTION_MAP (g_simple_action_group_new ());
-  g_action_map_add_action_entries (group, actions, G_N_ELEMENTS (actions), self);
-  gtk_widget_insert_action_group (GTK_WIDGET (tree), "grep", G_ACTION_GROUP (group));
+  pane = gtk_widget_get_ancestor (GTK_WIDGET (tree), IDE_TYPE_PANE);
+  g_assert (IDE_IS_PANE (pane));
+
+  self->group = g_simple_action_group_new ();
+  g_action_map_add_action_entries (G_ACTION_MAP (self->group),
+                                   actions,
+                                   G_N_ELEMENTS (actions),
+                                   self);
+  gtk_widget_insert_action_group (GTK_WIDGET (tree),
+                                  "grep",
+                                  G_ACTION_GROUP (self->group));
 }
 
 static void
@@ -103,7 +113,7 @@ gbp_grep_tree_addin_unload (IdeTreeAddin *addin,
   g_assert (IDE_IS_TREE_MODEL (model));
 
   gtk_widget_insert_action_group (GTK_WIDGET (tree), "grep", NULL);
-
+  g_clear_object (&self->group);
   self->tree = NULL;
 }
 
@@ -112,6 +122,7 @@ gbp_grep_tree_addin_selection_changed (IdeTreeAddin *addin,
                                        IdeTreeNode  *node)
 {
   GbpGrepTreeAddin *self = (GbpGrepTreeAddin *)addin;
+  GAction *action;
   gboolean enabled;
 
   g_assert (IDE_IS_MAIN_THREAD ());
@@ -119,10 +130,8 @@ gbp_grep_tree_addin_selection_changed (IdeTreeAddin *addin,
   g_assert (!node || IDE_IS_TREE_NODE (node));
 
   enabled = node && ide_tree_node_holds (node, IDE_TYPE_PROJECT_FILE);
-
-  dzl_gtk_widget_action_set (GTK_WIDGET (self->tree), "grep", "find-in-files",
-                             "enabled", enabled,
-                             NULL);
+  action = g_action_map_lookup_action (G_ACTION_MAP (self->group), "find-in-files");
+  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), enabled);
 }
 
 static void
@@ -134,7 +143,7 @@ tree_addin_iface_init (IdeTreeAddinInterface *iface)
 }
 
 G_DEFINE_FINAL_TYPE_WITH_CODE (GbpGrepTreeAddin, gbp_grep_tree_addin, G_TYPE_OBJECT,
-                         G_IMPLEMENT_INTERFACE (IDE_TYPE_TREE_ADDIN, tree_addin_iface_init))
+                               G_IMPLEMENT_INTERFACE (IDE_TYPE_TREE_ADDIN, tree_addin_iface_init))
 
 static void
 gbp_grep_tree_addin_class_init (GbpGrepTreeAddinClass *klass)
