@@ -20,9 +20,11 @@
 
 #define G_LOG_DOMAIN "xml-completion"
 
-#include <dazzle.h>
+#include "config.h"
+
 #include <gtksourceview/gtksource.h>
 #include <libpeas/peas.h>
+
 #include <libide-sourceview.h>
 
 #include "ide-xml-completion-attributes.h"
@@ -78,14 +80,14 @@ typedef struct
   IdeXmlRngDefine *define;
 } CompletionItem;
 
-static void      completion_provider_init (IdeCompletionProviderInterface *iface);
-static gboolean  process_matching_state   (MatchingState                  *state,
-                                           IdeXmlRngDefine                *define);
+static void      completion_provider_init (GtkSourceCompletionProviderInterface *iface);
+static gboolean  process_matching_state   (MatchingState                        *state,
+                                           IdeXmlRngDefine                      *define);
 
 G_DEFINE_FINAL_TYPE_WITH_CODE (IdeXmlCompletionProvider,
-                         ide_xml_completion_provider,
-                         IDE_TYPE_OBJECT,
-                         G_IMPLEMENT_INTERFACE (IDE_TYPE_COMPLETION_PROVIDER, completion_provider_init))
+                               ide_xml_completion_provider,
+                               IDE_TYPE_OBJECT,
+                               G_IMPLEMENT_INTERFACE (GTK_SOURCE_TYPE_COMPLETION_PROVIDER, completion_provider_init))
 
 static void
 populate_state_free (PopulateState *state)
@@ -119,14 +121,14 @@ copy_children (GPtrArray *children)
 }
 
 static void
-state_stack_item_free (gpointer *data)
+state_stack_item_clear (gpointer data)
 {
-  StateStackItem *item;
+  StateStackItem *item = data;
 
-  g_assert (data != NULL);
+  g_assert (item != NULL);
 
-  item = (StateStackItem *)data;
-  g_ptr_array_unref (item->children);
+  g_clear_pointer (&item->children, g_ptr_array_unref);
+  item->candidate_node = NULL;
 }
 
 static GArray *
@@ -135,7 +137,7 @@ state_stack_new (void)
   GArray *stack;
 
   stack = g_array_new (FALSE, TRUE, sizeof (StateStackItem));
-  g_array_set_clear_func (stack, (GDestroyNotify)state_stack_item_free);
+  g_array_set_clear_func (stack, state_stack_item_clear);
 
   return stack;
 }
@@ -373,7 +375,7 @@ completion_item_new (const gchar     *label,
 {
   CompletionItem *item;
 
-  g_assert (!dzl_str_empty0 (label));
+  g_assert (!ide_str_empty0 (label));
   g_assert (define != NULL);
 
   item = g_slice_new0 (CompletionItem);
@@ -505,7 +507,7 @@ is_define_equal_node (IdeXmlRngDefine  *define,
   g_assert (define != NULL);
   g_assert (IDE_IS_XML_SYMBOL_NODE (node));
 
-  return dzl_str_equal0 (ide_xml_symbol_node_get_element_name (node), define->name);
+  return ide_str_equal0 (ide_xml_symbol_node_get_element_name (node), define->name);
 }
 
 static gboolean
@@ -518,7 +520,7 @@ is_element_matching (MatchingState *state)
   g_assert (state->define->type == IDE_XML_RNG_DEFINE_ELEMENT);
 
   /* XXX: we skip element without a name for now */
-  if (dzl_str_empty0 ((gchar *)state->define->name))
+  if (ide_str_empty0 ((gchar *)state->define->name))
     return FALSE;
 
   if (state->children->len == 0)
@@ -530,7 +532,7 @@ is_element_matching (MatchingState *state)
   if (state->candidate_node == node)
     {
       name = (gchar *)state->define->name;
-      if (dzl_str_empty0 (state->prefix) || g_str_has_prefix ((gchar *)state->define->name, state->prefix))
+      if (ide_str_empty0 (state->prefix) || g_str_has_prefix ((gchar *)state->define->name, state->prefix))
         {
           state->candidate_node = NULL;
           state->retry = TRUE;
@@ -845,7 +847,7 @@ get_values_proposals (IdeXmlPosition  *position,
   GList *results = NULL;
 
   node = ide_xml_position_get_child_node (position);
-  g_assert (!dzl_str_empty0 (position->detail_name));
+  g_assert (!ide_str_empty0 (position->detail_name));
 
   if (NULL != (attributes = ide_xml_completion_attributes_get_matches (define, node, FALSE)))
     {
@@ -860,7 +862,7 @@ get_values_proposals (IdeXmlPosition  *position,
       for (gint j = 0; j < attributes->len; ++j)
         {
           match_item = g_ptr_array_index (attributes, j);
-          if (dzl_str_equal0 (detail_name, match_item->name))
+          if (ide_str_equal0 (detail_name, match_item->name))
             {
               attr_define = match_item->define;
               break;
@@ -934,7 +936,7 @@ populate_cb (GObject      *object,
   if (ide_task_return_error_if_cancelled (task))
     return;
 
-  proposals = g_list_store_new (IDE_TYPE_COMPLETION_PROPOSAL);
+  proposals = g_list_store_new (GTK_SOURCE_TYPE_COMPLETION_PROPOSAL);
 
   analysis = ide_xml_position_get_analysis (position);
   schemas = ide_xml_analysis_get_schemas (analysis);
@@ -1022,22 +1024,22 @@ cleanup:
 }
 
 static void
-ide_xml_completion_provider_populate_async (IdeCompletionProvider *provider,
-                                            IdeCompletionContext  *context,
+ide_xml_completion_provider_populate_async (GtkSourceCompletionProvider *provider,
+                                            GtkSourceCompletionContext  *context,
                                             GCancellable          *cancellable,
                                             GAsyncReadyCallback    callback,
                                             gpointer               user_data)
 {
   IdeXmlCompletionProvider *self = (IdeXmlCompletionProvider *)provider;
   g_autoptr(IdeTask) task = NULL;
-  GtkTextBuffer *buffer;
+  GtkSourceBuffer *buffer;
   IdeXmlService *service;
   PopulateState *state;
   IdeContext *ide_context;
   GtkTextIter iter;
 
   g_assert (IDE_IS_XML_COMPLETION_PROVIDER (self));
-  g_assert (IDE_IS_COMPLETION_CONTEXT (context));
+  g_assert (GTK_SOURCE_IS_COMPLETION_CONTEXT (context));
   g_assert (!cancellable || G_IS_CANCELLABLE (cancellable));
 
   task = ide_task_new (self, cancellable, callback, user_data);
@@ -1046,8 +1048,8 @@ ide_xml_completion_provider_populate_async (IdeCompletionProvider *provider,
   ide_context = ide_object_get_context (IDE_OBJECT (self));
   service = ide_xml_service_from_context (ide_context);
 
-  buffer = ide_completion_context_get_buffer (context);
-  ide_completion_context_get_bounds (context, &iter, NULL);
+  buffer = gtk_source_completion_context_get_buffer (context);
+  gtk_source_completion_context_get_bounds (context, &iter, NULL);
 
   state = g_slice_new0 (PopulateState);
   state->file = g_object_ref (ide_buffer_get_file (IDE_BUFFER (buffer)));
@@ -1067,7 +1069,7 @@ ide_xml_completion_provider_populate_async (IdeCompletionProvider *provider,
 }
 
 static GListModel *
-ide_xml_completion_provider_populate_finish (IdeCompletionProvider  *provider,
+ide_xml_completion_provider_populate_finish (GtkSourceCompletionProvider  *provider,
                                              GAsyncResult           *result,
                                              GError                **error)
 {
@@ -1077,9 +1079,9 @@ ide_xml_completion_provider_populate_finish (IdeCompletionProvider  *provider,
   return ide_task_propagate_object (IDE_TASK (result), error);
 }
 
-static gboolean
-ide_xml_completion_provider_refilter (IdeCompletionProvider *provider,
-                                      IdeCompletionContext  *context,
+static void
+ide_xml_completion_provider_refilter (GtkSourceCompletionProvider *provider,
+                                      GtkSourceCompletionContext  *context,
                                       GListModel            *model)
 {
   g_autofree gchar *word = NULL;
@@ -1087,11 +1089,11 @@ ide_xml_completion_provider_refilter (IdeCompletionProvider *provider,
   guint n_items;
 
   g_assert (IDE_IS_XML_COMPLETION_PROVIDER (provider));
-  g_assert (IDE_IS_COMPLETION_CONTEXT (context));
+  g_assert (GTK_SOURCE_IS_COMPLETION_CONTEXT (context));
   g_assert (G_IS_LIST_MODEL (model));
 
   n_items = g_list_model_get_n_items (model);
-  word = ide_completion_context_get_word (context);
+  word = gtk_source_completion_context_get_word (context);
   casefold = g_utf8_casefold (word, -1);
 
   /* TODO: This doesn't work with backspace. We need to track things
@@ -1104,56 +1106,65 @@ ide_xml_completion_provider_refilter (IdeCompletionProvider *provider,
       const gchar *label = ide_xml_proposal_get_label (proposal);
       guint priority;
 
-      if (!ide_completion_fuzzy_match (label, casefold, &priority))
+      if (!gtk_source_completion_fuzzy_match (label, casefold, &priority))
         g_list_store_remove (G_LIST_STORE (model), i - 1);
     }
-
-  return TRUE;
 }
 
 static void
-ide_xml_completion_provider_display_proposal (IdeCompletionProvider   *provider,
-                                              IdeCompletionListBoxRow *row,
-                                              IdeCompletionContext    *context,
-                                              const gchar             *typed_text,
-                                              IdeCompletionProposal   *proposal)
+ide_xml_completion_provider_display (GtkSourceCompletionProvider *provider,
+                                     GtkSourceCompletionContext  *context,
+                                     GtkSourceCompletionProposal *proposal,
+                                     GtkSourceCompletionCell     *cell)
 {
   IdeXmlProposal *item = (IdeXmlProposal *)proposal;
-  const gchar *label;
 
   g_assert (IDE_IS_XML_COMPLETION_PROVIDER (provider));
-  g_assert (IDE_IS_COMPLETION_LIST_BOX_ROW (row));
-  g_assert (IDE_IS_COMPLETION_CONTEXT (context));
-  g_assert (IDE_IS_COMPLETION_PROPOSAL (proposal));
+  g_assert (GTK_SOURCE_IS_COMPLETION_CELL (cell));
+  g_assert (GTK_SOURCE_IS_COMPLETION_CONTEXT (context));
+  g_assert (GTK_SOURCE_IS_COMPLETION_PROPOSAL (proposal));
 
-  label = ide_xml_proposal_get_label (item);
+  switch (gtk_source_completion_cell_get_column (cell))
+    {
+    case GTK_SOURCE_COMPLETION_COLUMN_ICON:
+      gtk_source_completion_cell_set_icon_name (cell, "text-xml-symbolic");
+      break;
 
-  ide_completion_list_box_row_set_icon_name (row, NULL);
-  ide_completion_list_box_row_set_left (row, NULL);
-  ide_completion_list_box_row_set_right (row, NULL);
-  ide_completion_list_box_row_set_center_markup (row, label);
+    case GTK_SOURCE_COMPLETION_COLUMN_TYPED_TEXT:
+      gtk_source_completion_cell_set_markup (cell, ide_xml_proposal_get_label (item));
+      break;
+
+    case GTK_SOURCE_COMPLETION_COLUMN_COMMENT:
+    case GTK_SOURCE_COMPLETION_COLUMN_DETAILS:
+    case GTK_SOURCE_COMPLETION_COLUMN_BEFORE:
+    case GTK_SOURCE_COMPLETION_COLUMN_AFTER:
+    default:
+      gtk_source_completion_cell_set_text (cell, NULL);
+      break;
+    }
 }
 
 static void
-ide_xml_completion_provider_activate_proposal (IdeCompletionProvider *provider,
-                                               IdeCompletionContext  *context,
-                                               IdeCompletionProposal *proposal,
-                                               const GdkEventKey     *key)
+ide_xml_completion_provider_activate (GtkSourceCompletionProvider *provider,
+                                      GtkSourceCompletionContext  *context,
+                                      GtkSourceCompletionProposal *proposal)
 {
   IdeXmlProposal *item = (IdeXmlProposal *)proposal;
+  GtkSourceBuffer *source_buffer;
   GtkTextBuffer *buffer;
   GtkTextIter begin, end;
   const gchar *text;
 
   g_assert (IDE_IS_XML_COMPLETION_PROVIDER (provider));
-  g_assert (IDE_IS_COMPLETION_CONTEXT (context));
-  g_assert (IDE_IS_COMPLETION_PROPOSAL (proposal));
+  g_assert (GTK_SOURCE_IS_COMPLETION_CONTEXT (context));
+  g_assert (GTK_SOURCE_IS_COMPLETION_PROPOSAL (proposal));
 
   text = ide_xml_proposal_get_text (item);
 
-  buffer = ide_completion_context_get_buffer (context);
-  ide_completion_context_get_bounds (context, &begin, &end);
+  source_buffer = gtk_source_completion_context_get_buffer (context);
+  gtk_source_completion_context_get_bounds (context, &begin, &end);
 
+  buffer = GTK_TEXT_BUFFER (source_buffer);
   gtk_text_buffer_begin_user_action (buffer);
   gtk_text_buffer_delete (buffer, &begin, &end);
   gtk_text_buffer_insert (buffer, &begin, text, -1);
@@ -1180,10 +1191,10 @@ ide_xml_completion_provider_init (IdeXmlCompletionProvider *self)
 }
 
 static void
-completion_provider_init (IdeCompletionProviderInterface *iface)
+completion_provider_init (GtkSourceCompletionProviderInterface *iface)
 {
-  iface->activate_proposal = ide_xml_completion_provider_activate_proposal;
-  iface->display_proposal = ide_xml_completion_provider_display_proposal;
+  iface->activate = ide_xml_completion_provider_activate;
+  iface->display = ide_xml_completion_provider_display;
   iface->populate_async = ide_xml_completion_provider_populate_async;
   iface->populate_finish = ide_xml_completion_provider_populate_finish;
   iface->refilter = ide_xml_completion_provider_refilter;
